# This file was auto-generated by Fern from our API Definition.

import typing
from .cost_function import CostFunction
from .parameter_constraint import ParameterConstraint

Statement = typing.Union[CostFunction, ParameterConstraint]



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import pydantic
import typing
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class ExtractedPoint(UniversalBaseModel):
    """
    Represents a single extracted point from the plot.
    """

    value_x: float = pydantic.Field()
    """
    X-axis value of the extracted point.
    """

    value_y: float = pydantic.Field()
    """
    Y-axis value of the extracted point.
    """

    img_coord_x: typing.Optional[float] = pydantic.Field(default=None)
    """
    X coordinate in the image.
    """

    img_coord_y: typing.Optional[float] = pydantic.Field(default=None)
    """
    Y coordinate in the image.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from ... import core
from ...core.request_options import RequestOptions
from ...types.plot_parser_output import PlotParserOutput
from ...core.pydantic_utilities import parse_obj_as
from ...errors.unprocessable_entity_error import UnprocessableEntityError
from ...types.http_validation_error import HttpValidationError
from json.decoder import JSONDecodeError
from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class PlotClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def points(
        self,
        *,
        plot_img: core.File,
        method: typing.Optional[int] = None,
        plot_info: typing.Optional[str] = None,
        get_img_coords: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PlotParserOutput:
        """
        Extracts points from plots

        Parameters
        ----------
        plot_img : core.File
            See core.File for more documentation

        method : typing.Optional[int]
            Can specify a specific method to extract points

        plot_info : typing.Optional[str]
            Can add specific plot info

        get_img_coords : typing.Optional[bool]
            Whether to get coords of points on image

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PlotParserOutput
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.document.plot.points()
        """
        _response = self._client_wrapper.httpx_client.request(
            "document/plot/points",
            method="POST",
            params={
                "method": method,
                "plot_info": plot_info,
                "get_img_coords": get_img_coords,
            },
            data={},
            files={
                "plot_img": plot_img,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    PlotParserOutput,
                    parse_obj_as(
                        type_=PlotParserOutput,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncPlotClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def points(
        self,
        *,
        plot_img: core.File,
        method: typing.Optional[int] = None,
        plot_info: typing.Optional[str] = None,
        get_img_coords: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PlotParserOutput:
        """
        Extracts points from plots

        Parameters
        ----------
        plot_img : core.File
            See core.File for more documentation

        method : typing.Optional[int]
            Can specify a specific method to extract points

        plot_info : typing.Optional[str]
            Can add specific plot info

        get_img_coords : typing.Optional[bool]
            Whether to get coords of points on image

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PlotParserOutput
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document.plot.points()


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "document/plot/points",
            method="POST",
            params={
                "method": method,
                "plot_info": plot_info,
                "get_img_coords": get_img_coords,
            },
            data={},
            files={
                "plot_img": plot_img,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    PlotParserOutput,
                    parse_obj_as(
                        type_=PlotParserOutput,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from . import constraints, solution
from .solution import SolutionBodyValuesValue

__all__ = ["SolutionBodyValuesValue", "constraints", "solution"]



# This file was auto-generated by Fern from our API Definition.

import typing
from ..core.client_wrapper import SyncClientWrapper
from .data_files.client import DataFilesClient
from ..types.user_requirement import UserRequirement
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..core.pydantic_utilities import parse_obj_as
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.http_validation_error import HttpValidationError
from json.decoder import JSONDecodeError
from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper
from .data_files.client import AsyncDataFilesClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RequirementsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper
        self.data_files = DataFilesClient(client_wrapper=self._client_wrapper)

    def check(
        self, *, request: typing.Sequence[UserRequirement], request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request : typing.Sequence[UserRequirement]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic, UserRequirement

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.requirements.check(
            request=[
                UserRequirement(
                    latex_symbol="latex_symbol",
                    requirement_name="requirement_name",
                    tolerance=1.1,
                    value=1.1,
                    units="units",
                )
            ],
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "requirements/check",
            method="POST",
            json=convert_and_respect_annotation_metadata(
                object_=request, annotation=typing.Sequence[UserRequirement], direction="write"
            ),
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    typing.Optional[typing.Any],
                    parse_obj_as(
                        type_=typing.Optional[typing.Any],  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncRequirementsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper
        self.data_files = AsyncDataFilesClient(client_wrapper=self._client_wrapper)

    async def check(
        self, *, request: typing.Sequence[UserRequirement], request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request : typing.Sequence[UserRequirement]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic, UserRequirement

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.requirements.check(
                request=[
                    UserRequirement(
                        latex_symbol="latex_symbol",
                        requirement_name="requirement_name",
                        tolerance=1.1,
                        value=1.1,
                        units="units",
                    )
                ],
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "requirements/check",
            method="POST",
            json=convert_and_respect_annotation_metadata(
                object_=request, annotation=typing.Sequence[UserRequirement], direction="write"
            ),
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    typing.Optional[typing.Any],
                    parse_obj_as(
                        type_=typing.Optional[typing.Any],  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



import ipywidgets as widgets  # type: ignore
from IPython.display import display, Math, HTML  # type: ignore
import json  # type: ignore
import os
import hypernetx as hnx  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import re
from dataclasses import dataclass, asdict


@dataclass
class RequirementUserInput:
    requirement_name: str
    latex_symbol: str
    value: int
    units: str
    tolerance: float


def _find_symbol(name, variable_dict):
    matching_keys = [key for key, value in variable_dict.items() if name in value["name"]]

    if not matching_keys:
        matching_keys.append("unknown")

    return matching_keys[0]


def requirements_from_table(results, variable_dict):
    requirements = []

    for key, value in results["values"].items():
        latex_symbol = _find_symbol(key, variable_dict)

        name = key
        numerical_value = value["Value"]
        unit = value["Units"]

        requirements.append(
            RequirementUserInput(
                requirement_name=name,
                latex_symbol=latex_symbol,
                value=numerical_value,
                units=unit,
                tolerance=0.0,
            )
        )

    return requirements


def interactive_table(variable_dict, file_path="./custom_presets.json"):
    """
    Creates an interactive table for IMAGING_TELESCOPE,
    PAYLOAD, and user-defined custom templates.
    Adds or deletes rows, and can save custom templates persistently in JSON.

    Parameters
    ----------
    variable_dict : dict
        Dictionary used to populate the "Add Requirement" dropdown, e.g.:
          {
            "var_key1": {"name": "Human-readable variable name1"},
            "var_key2": {"name": "Human-readable variable name2"},
            ...
          }
    file_path : str, optional
        JSON file path where we load and save user-created custom templates.

    Returns
    -------
    dict
        Contains user inputs after pressing "Submit" button.
    """

    # ---------------------------------------------------------------
    # 1) Define built-in templates and units directly inside the function
    # ---------------------------------------------------------------
    IMAGING_TELESCOPE_template = {
        "Resolution (panchromatic)": 0,
        "Ground sampling distance (panchromatic)": 0,
        "Resolution (multispectral)": 0,
        "Ground sampling distance (multispectral)": 0,
        "Altitude": 0,
        "Half field of view": 0,
        "Mirror aperture": 0,
        "F-number": 0,
        "Focal length": 0,
        "Pixel size (panchromatic)": 0,
        "Pixel size (multispectral)": 0,
        "Swath width": 0,
    }

    IMAGING_TELESCOPE = {
        "Resolution (panchromatic)": 1.23529,
        "Ground sampling distance (panchromatic)": 0.61765,
        "Resolution (multispectral)": 1.81176,
        "Ground sampling distance (multispectral)": 0.90588,
        "Altitude": 420000,
        "Half field of view": 0.017104227,
        "Mirror aperture": 0.85,
        "F-number": 6.0,
        "Focal length": 5.1,
        "Pixel size (panchromatic)": 7.5e-6,
        "Pixel size (multispectral)": 11e-6,
        "Swath width": 14368.95,
    }

    IMAGING_TELESCOPE_UNITS = {
        "Resolution (panchromatic)": "m",
        "Ground sampling distance (panchromatic)": "m",
        "Resolution (multispectral)": "m",
        "Ground sampling distance (multispectral)": "m",
        "Altitude": "m",
        "Half field of view": "rad",
        "Mirror aperture": "m",
        "F-number": "dimensionless",
        "Focal length": "m",
        "Pixel size (panchromatic)": "m",
        "Pixel size (multispectral)": "m",
        "Swath width": "m",
    }

    PAYLOAD_1 = {
        "Resolution (panchromatic)": 15.4,
        "Ground sampling distance (panchromatic)": 7.7,
        "Resolution (multispectral)": 0.0,
        "Ground sampling distance (multispectral)": 0.0,
        "Altitude": 420000,
        "Half field of view": 0.005061455,
        "Mirror aperture": 0.85,
        "F-number": 1.0,
        "Focal length": 0.3,
        "Pixel size (panchromatic)": 5.5e-6,
        "Swath width": 4251.66,
    }

    # ---------------------------------------------------------------
    # 2) Create a preset_options_dict with built-in templates
    # ---------------------------------------------------------------
    preset_options_dict = {
        "Select a template": [],
        "IMAGING TELESCOPE": list(IMAGING_TELESCOPE.keys()),
        "IMAGING TELESCOPE template": list(IMAGING_TELESCOPE_template.keys()),
        "PAYLOAD": list(PAYLOAD_1.keys()),
    }

    # ---------------------------------------------------------------
    # 3) Helper functions for loading/saving custom presets from JSON
    # ---------------------------------------------------------------
    def load_custom_presets(file_path):
        if os.path.exists(file_path):
            with open(file_path, "r") as f:
                return json.load(f)
        return {}

    def save_custom_presets(custom_data, file_path):
        with open(file_path, "w") as f:
            json.dump(custom_data, f, indent=2)

    # ---------------------------------------------------------------
    # 4) Load custom presets from JSON (if any) and integrate them
    # ---------------------------------------------------------------
    custom_presets = load_custom_presets(file_path)

    for custom_name, values_dict in custom_presets.items():
        preset_options_dict[custom_name] = list(values_dict.keys())

    # ---------------------------------------------------------------
    # 5) For the "Add Requirement" dropdown
    # ---------------------------------------------------------------
    variable_names = [details["name"] for details in variable_dict.values()]

    # This dict will store the final user inputs after pressing "Submit"
    result = {}

    # Main dropdown to pick a template
    dropdown = widgets.Dropdown(
        options=list(preset_options_dict.keys()),
        description="Select Option:",
        style={"description_width": "initial"},
    )

    # Container for all rows (including the header)
    rows_output = widgets.VBox()
    # Container for user messages
    message_output = widgets.Output()

    # We'll dynamically resize this label width
    name_label_width = ["150px"]

    # Dictionary to keep track of row widget references
    value_widgets = {}

    # ---------------------------------------------------------------
    # 6) display_table(change): Re-populate rows when user selects a template
    # ---------------------------------------------------------------
    def display_table(change):
        selected_option = change["new"]

        # Clear existing rows in the GUI
        rows_output.children = ()
        value_widgets.clear()

        if selected_option in preset_options_dict:
            rows = preset_options_dict[selected_option]

            if selected_option != "Select a template" and len(rows) > 0:
                max_name_length = max(len(r) for r in rows)
                name_label_width[0] = f"{max_name_length + 2}ch"
            else:
                name_label_width[0] = "42ch"

            # Create a header row
            header_labels = [
                widgets.Label(
                    value="Name",
                    layout=widgets.Layout(width=name_label_width[0]),
                    style={"font_weight": "bold"},
                ),
                widgets.Label(
                    value="Value",
                    layout=widgets.Layout(width="150px"),
                    style={"font_weight": "bold"},
                ),
                widgets.Label(
                    value="Units",
                    layout=widgets.Layout(width="150px"),
                    style={"font_weight": "bold"},
                ),
            ]
            header = widgets.HBox(header_labels)
            header.layout = widgets.Layout(
                border="1px solid black",
                padding="5px",
            )
            rows_output.children += (header,)

            for row_name in rows:
                # Figure out default values if it's one of our built-ins
                if selected_option == "IMAGING TELESCOPE":
                    default_value = IMAGING_TELESCOPE.get(row_name, 0.0)
                    default_unit = IMAGING_TELESCOPE_UNITS.get(row_name, "")
                elif selected_option == "PAYLOAD":
                    default_value = PAYLOAD_1.get(row_name, 0.0)
                    default_unit = IMAGING_TELESCOPE_UNITS.get(row_name, "")
                elif selected_option in custom_presets:
                    default_value = custom_presets[selected_option].get(row_name, {}).get("Value", 0.0)
                    default_unit = custom_presets[selected_option].get(row_name, {}).get("Units", "")
                else:
                    default_value = 0.0
                    default_unit = ""

                name_label = widgets.Label(
                    value=row_name,
                    layout=widgets.Layout(width=name_label_width[0]),
                )
                value_text = widgets.FloatText(
                    value=default_value,
                    layout=widgets.Layout(width="150px"),
                )
                units_text = widgets.Text(
                    value=default_unit,
                    layout=widgets.Layout(width="150px"),
                )

                row = widgets.HBox([name_label, value_text, units_text])
                value_widgets[row_name] = row
                rows_output.children += (row,)

    dropdown.observe(display_table, names="value")

    # Display the UI
    display(dropdown)
    display(rows_output)
    display(message_output)

    # ---------------------------------------------------------------
    # 7) submit_values(): Gather the current table's values into `result`
    # ---------------------------------------------------------------
    def submit_values(_):
        updated_values = {}
        for k, widget in value_widgets.items():
            label_or_variable = widget.children[0].value
            val = widget.children[1].value
            unit = widget.children[2].value
            updated_values[label_or_variable] = {"Value": val, "Units": unit}

        result["values"] = updated_values

    # ---------------------------------------------------------------
    # 8) add_req(): Adds a new, blank row to the bottom
    # ---------------------------------------------------------------
    def add_req(_):
        unique_key = f"req_{len([kk for kk in value_widgets if kk.startswith('req_')]) + 1}"

        variable_dropdown = widgets.Dropdown(
            options=variable_names,
            description="Variable:",
            style={"description_width": "initial"},
            layout=widgets.Layout(width=str(50) + "ch"),
        )
        value_text = widgets.FloatText(
            placeholder="Value",
            layout=widgets.Layout(width="150px"),
        )
        units_text = widgets.Text(
            placeholder="Units",
            layout=widgets.Layout(width="150px"),
        )

        new_row = widgets.HBox([variable_dropdown, value_text, units_text])
        rows_output.children += (new_row,)
        value_widgets[unique_key] = new_row

    # ---------------------------------------------------------------
    # 9) delete_req(): Delete the last row (if there's more than the header)
    # ---------------------------------------------------------------
    def delete_req(_):
        with message_output:
            message_output.clear_output()
            if len(rows_output.children) > 1:
                children_list = list(rows_output.children)
                last_row = children_list.pop()  # remove from display
                rows_output.children = tuple(children_list)

                # remove from the dictionary
                for k in reversed(list(value_widgets.keys())):
                    if value_widgets[k] is last_row:
                        del value_widgets[k]
                        break
            else:
                print("No row available to delete.")

    # ---------------------------------------------------------------
    # 10) save_requirements():
    #   - Gathers rows,
    #   - Creates a new "custom_n" entry in preset_options_dict,
    #   - Also updates custom_presets + JSON file,
    #   - So it persists across restarts.
    # ---------------------------------------------------------------
    custom_count = len([k for k in preset_options_dict if k.startswith("Custom-")])

    def save_requirements(_):
        nonlocal custom_count
        custom_count += 1
        new_option_name = f"custom_{custom_count}"

        # Gather current row data
        updated_values = {}
        for key, widget in value_widgets.items():
            row_label = widget.children[0].value
            val = widget.children[1].value
            unit = widget.children[2].value
            updated_values[row_label] = {"Value": val, "Units": unit}

        # Get row names for the new template
        new_template_rows = list(updated_values.keys())

        # Insert new key into preset_options_dict so it appears in dropdown
        preset_options_dict[new_option_name] = new_template_rows

        # Store the data in custom_presets
        custom_presets[new_option_name] = updated_values

        # Persist to JSON
        save_custom_presets(custom_presets, file_path)

        # Update the dropdown
        dropdown.options = list(preset_options_dict.keys())

    # ---------------------------------------------------------------
    # 11) Create & display the buttons
    # ---------------------------------------------------------------
    submit_button = widgets.Button(description="Submit", button_style="success")
    submit_button.on_click(submit_values)

    add_req_button = widgets.Button(description="Add Requirement", button_style="primary")
    add_req_button.on_click(add_req)

    del_req_button = widgets.Button(description="Delete Requirement", button_style="danger")
    del_req_button.on_click(delete_req)

    save_req_button = widgets.Button(description="Save", button_style="info")
    save_req_button.on_click(save_requirements)

    buttons_box = widgets.HBox([submit_button, add_req_button, del_req_button, save_req_button])
    display(buttons_box)

    return result


def _get_node_names_for_node_lables(node_labels, api_requirements):
    # Create the output list
    node_names = []

    # Iterate through each symbol in S
    for symbol in node_labels:
        # Search for the matching requirement
        symbol = symbol.replace("$", "")
        for req in api_requirements:
            if req["latex_symbol"] == symbol:
                # Add the matching tuple to SS
                node_names.append((req["latex_symbol"], req["requirement_name"]))
                break  # Stop searching once a match is found

    return node_names


def _get_latex_string_format(input_string):
    """
    Properly formats LaTeX strings for matplotlib when text.usetex is False.
    No escaping needed since mathtext handles backslashes properly.
    """
    return f"${input_string}$"  # No backslash escaping required


def _get_requirements_set(requirements):
    variable_set = set()
    for req in requirements:
        variable_set.add(req["latex_symbol"])

    return variable_set


def _find_vars_in_eq(equation, variable_set):
    patterns = [re.escape(var) for var in variable_set]
    combined_pattern = r"|".join(patterns)
    matches = re.findall(combined_pattern, equation)
    return {rf"${match}$" for match in matches}


def _add_used_vars_to_results(api_results, api_requirements):
    requirements = _get_requirements_set(api_requirements)

    for key, value in api_results["results"].items():
        latex_equation = value.get("latex_equation")
        # print(latex_equation)
        if latex_equation:
            used_vars = _find_vars_in_eq(latex_equation, requirements)
            api_results["results"][key]["used_vars"] = used_vars

    return api_results


def display_full_results(equations_dict, requirements=None, show_hypergraph=True):
    """Display equation validation results optimized for dark theme notebooks."""
    results = equations_dict.get("results", {})

    def format_equation(latex_eq):
        inner = latex_eq[3:-1]
        lhs, rhs = inner.split(",", 1)
        return f"{lhs} = {rhs}"

    matching = []
    non_matching = []

    for key, value in results.items():
        equation_data = {
            "latex": format_equation(value.get("latex_equation")),
            "lhs": value.get("lhs"),
            "rhs": value.get("rhs"),
            "diff": abs(value.get("lhs", 0) - value.get("rhs", 0)),
            "percent_diff": abs(value.get("lhs", 0) - value.get("rhs", 0)) / max(abs(value.get("rhs", 0)), 1e-10) * 100,
        }
        if value.get("match"):
            matching.append(equation_data)
        else:
            non_matching.append(equation_data)

    # Summary header with dark theme
    total = len(results)
    display(
        HTML(
            '<div style="background-color:#1e1e1e; padding:20px; border-radius:10px; margin:20px 0; '
            'border:1px solid #3e3e3e;">'
            f'<h2 style="font-family:Arial; color:#e0e0e0; margin-bottom:15px">Equation Validation Analysis</h2>'
            f'<p style="font-family:Arial; font-size:16px; color:#e0e0e0">'
            f"<b>Total equations analyzed:</b> {total}<br>"
            f'<span style="color:#4caf50">✅ Matching equations: {len(matching)}</span><br>'
            f'<span style="color:#ff5252">❌ Non-matching equations: {len(non_matching)}</span></p>'
            "</div>"
        )
    )

    # Non-matching equations
    if non_matching:
        display(
            HTML(
                '<div style="background-color:#2d1f1f; padding:20px; border-radius:10px; margin:20px 0; '
                'border:1px solid #4a2f2f;">'
                '<h3 style="color:#ff5252; font-family:Arial">⚠️ Equations Not Satisfied</h3>'
            )
        )

        for eq in non_matching:
            display(Math(eq["latex"]))
            display(
                HTML(
                    '<div style="font-family:monospace; margin-left:20px; margin-bottom:20px; '
                    "background-color:#2a2a2a; color:#e0e0e0; padding:15px; border-radius:5px; "
                    'border-left:4px solid #ff5252">'
                    f"Left side  = {eq['lhs']:.6g}<br>"
                    f"Right side = {eq['rhs']:.6g}<br>"
                    f"Absolute difference = {eq['diff']:.6g}<br>"
                    f"Relative difference = {eq['percent_diff']:.2f}%"
                    "</div>"
                )
            )

        display(HTML("</div>"))

    # Matching equations
    if matching:
        display(
            HTML(
                '<div style="background-color:#1f2d1f; padding:20px; border-radius:10px; margin:20px 0; '
                'border:1px solid #2f4a2f;">'
                '<h3 style="color:#4caf50; font-family:Arial">✅ Satisfied Equations</h3>'
            )
        )

        for eq in matching:
            display(Math(eq["latex"]))
            display(
                HTML(
                    '<div style="font-family:monospace; margin-left:20px; margin-bottom:20px; '
                    "background-color:#2a2a2a; color:#e0e0e0; padding:15px; border-radius:5px; "
                    'border-left:4px solid #4caf50">'
                    f"Value = {eq['lhs']:.6g}"
                    "</div>"
                )
            )

        display(HTML("</div>"))

    # Hypergraph visualization
    if show_hypergraph and requirements:
        display(
            HTML(
                '<div style="background-color:#1e1e1e; padding:20px; border-radius:10px; margin:20px 0; '
                'border:1px solid #3e3e3e;">'
                '<h3 style="color:#e0e0e0; font-family:Arial">🔍 Equation Relationship Analysis</h3>'
                '<p style="font-family:Arial; color:#e0e0e0">The following graph shows how variables are connected through equations:</p>'
                "</div>"
            )
        )

        list_api_requirements = [asdict(req) for req in requirements]

        # Match get_eq_hypergraph settings exactly
        plt.rcParams["text.usetex"] = False
        plt.rcParams["mathtext.fontset"] = "stix"
        plt.rcParams["font.family"] = "serif"

        equations_dict = _add_used_vars_to_results(equations_dict, list_api_requirements)

        # Prepare hypergraph data
        hyperedges = {}
        for eq, details in equations_dict["results"].items():
            hyperedges[_get_latex_string_format(details["latex_equation"])] = details["used_vars"]

        # Create and plot the hypergraph
        H = hnx.Hypergraph(hyperedges)
        plt.figure(figsize=(16, 12))

        # Draw hypergraph with exact same settings as get_eq_hypergraph
        hnx.draw(
            H,
            with_edge_labels=True,
            edge_labels_on_edge=False,
            node_labels_kwargs={"fontsize": 14},
            edge_labels_kwargs={"fontsize": 14},
            layout_kwargs={"seed": 42, "scale": 2.5},
        )

        node_labels = list(H.nodes)
        symbol_explanations = _get_node_names_for_node_lables(node_labels, list_api_requirements)

        explanation_text = "\n".join([f"${symbol}$: {desc}" for symbol, desc in symbol_explanations])
        plt.annotate(
            explanation_text,
            xy=(1.05, 0.5),
            xycoords="axes fraction",
            fontsize=14,
            verticalalignment="center",
        )

        plt.title(r"Enhanced Hypergraph of Equations and Variables", fontsize=20)
        plt.show()

    return None


def get_numerical_values(ax_client, path, constants_of_interest):
    with open(path, "rb") as f:
        file = f.read()

    constants = ax_client.document.constants(file=file, constants=constants_of_interest).constants
    print(constants)
    # Create a dictionary to store processed values
    processed_values = {}

    # Process each constant name from the constants dictionary
    for constant_name in constants:
        value_str = constants[constant_name]  # Get the value directly from the dictionary

        if value_str is None:
            # Handle None values
            processed_values[constant_name] = {"Value": 0.0, "Units": "unknown"}
        elif "F/" in value_str:
            # Handle F-number values
            f_number = float(value_str.split("/")[-1])
            processed_values[constant_name] = {"Value": f_number, "Units": "dimensionless"}
        elif "f/" in value_str:
            # Handle F-number values
            f_number = float(value_str.split("/")[-1])
            processed_values[constant_name] = {"Value": f_number, "Units": "dimensionless"}
        else:
            # Handle normal values with units
            # Split on the last space to separate value and unit
            parts = value_str.rsplit(" ", 1)
            if len(parts) == 2:
                value, unit = parts
                # Convert value to float
                value = float(value)

                # Handle unit conversions to meters
                if unit == "\u00b5m":  # micrometer
                    value *= 1e-6
                    unit = "m"
                elif unit == "mm":  # millimeter
                    value *= 1e-3
                    unit = "m"
                elif unit == "km":  # kilometer
                    value *= 1e3
                    unit = "m"
                # Handle degree conversions to radians
                elif unit in ["deg", "°", "degree", "degrees"]:  # degrees
                    value = value * (3.14159265359 / 180.0)  # convert to radians
                    unit = "rad"

                processed_values[constant_name] = {"Value": value, "Units": unit}
            else:
                # If no unit is found
                processed_values[constant_name] = {"Value": float(parts[0]), "Units": "unknown"}

    # Save as custom preset
    filename = os.path.basename(path)
    with open("./custom_presets.json", "r+") as f:
        presets = json.load(f)
        presets[filename] = processed_values
        f.seek(0)
        json.dump(presets, f, indent=2)
        f.truncate()

    return processed_values



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .port_instance_layout_info import PortInstanceLayoutInfo
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class ComponentInstanceLayoutInfo(UniversalBaseModel):
    """
    Class to represent layout information of a component in a circuit.
    This information is only obtained from the code, not from the PDK.

    Note that this data is NOT the same as the placement data in the Netlist.
    Note that this data is NOT the same as the component_info data in the Netlist.
    """

    component: str
    center_x: float
    center_y: float
    size_x: float
    size_y: float
    ports: typing.Dict[str, PortInstanceLayoutInfo]

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import base64
import dill  # type: ignore
import json
import requests  # type: ignore
import os
import time
from typing import Dict, Optional, Sequence

from .base_client import BaseClient, AsyncBaseClient
from . import ParseResponse, EquationProcessingResponse
from .axtract.axtract_report import create_report
from .axtract.validation_results import display_full_results
from .types.variable_requirement import VariableRequirement
from .types.equation_validation_result import EquationValidationResult


class Axiomatic(BaseClient):
    def __init__(self, *args, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = 1200
        super().__init__(*args, **kwargs)

        self.document_helper = DocumentHelper(self)
        self.tools_helper = ToolsHelper(self)
        self.axtract_helper = AxtractHelper(self)


class DocumentHelper:
    _ax_client: Axiomatic

    def __init__(self, ax_client: Axiomatic):
        self._ax_client = ax_client

    def pdf_from_url(self, url: str) -> ParseResponse:
        """Download a PDF document from a URL and parse it into a Markdown response."""
        if "arxiv" in url and "abs" in url:
            url = url.replace("abs", "pdf")
            print("The URL is an arXiv abstract page. Replacing 'abs' with 'pdf' to download the PDF.")
        file = requests.get(url)
        response = self._ax_client.document.parse(file=file.content)
        return response

    def pdf_from_file(self, path: str) -> ParseResponse:
        """Open a PDF document from a file path and parse it into a Markdown response."""
        with open(path, "rb") as f:
            file = f.read()
        response = self._ax_client.document.parse(file=file)
        return response

    def plot_b64_images(self, images: Dict[str, str]):
        """Plot a dictionary of base64 images."""
        import ipywidgets as widgets  # type: ignore
        from IPython.display import display  # type: ignore

        base64_images = list(images.values())
        current_index = [0]

        def display_base64_image(index):
            image_widget.value = base64.b64decode(base64_images[index])

        def navigate_image(change):
            current_index[0] = (current_index[0] + change) % len(base64_images)
            display_base64_image(current_index[0])

        image_widget = widgets.Image(format="png", width=600)
        prev_button = widgets.Button(description="Previous", icon="arrow-left")
        next_button = widgets.Button(description="Next", icon="arrow-right")

        prev_button.on_click(lambda b: navigate_image(-1))
        next_button.on_click(lambda b: navigate_image(1))

        buttons = widgets.HBox([prev_button, next_button])
        layout = widgets.VBox([buttons, image_widget])

        display(layout)
        display_base64_image(current_index[0])

    def save_parsed_pdf(self, response: ParseResponse, path: str):
        """Save a parsed PDF response to a file."""
        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, "text.md"), "w") as f:
            f.write(response.markdown)

        if response.images:
            for img_name, img in response.images.items():
                with open(os.path.join(path, f"{img_name}.png"), "wb") as f:
                    f.write(base64.b64decode(img))

        with open(os.path.join(path, "interline_equations.json"), "w") as f:
            json.dump(response.interline_equations, f)

        with open(os.path.join(path, "inline_equations.json"), "w") as f:
            json.dump(response.inline_equations, f)

    def load_parsed_pdf(self, path: str) -> ParseResponse:
        """Load a parsed PDF response from a file."""
        with open(os.path.join(path, "text.md"), "r") as f:
            markdown = f.read()

        images = {}
        for img_name in os.listdir(path):
            if img_name.endswith((".png")):
                with open(os.path.join(path, img_name), "rb") as img_file:
                    images[img_name] = base64.b64encode(img_file.read()).decode("utf-8")

        with open(os.path.join(path, "interline_equations.json"), "r") as f:
            interline_equations = json.load(f)

        with open(os.path.join(path, "inline_equations.json"), "r") as f:
            inline_equations = json.load(f)

        return ParseResponse(
            markdown=markdown,
            images=images,
            interline_equations=interline_equations,
            inline_equations=inline_equations,
        )


class AxtractHelper:
    _ax_client: Axiomatic

    def __init__(self, ax_client: Axiomatic):
        """Initialize the AxtractHelper with an Axiomatic client.

        Args:
            ax_client (Axiomatic): The Axiomatic client instance to use for API calls
        """
        self._ax_client = ax_client

    def create_report(self, response: EquationProcessingResponse, path: str):
        """Create a report from equation extraction results.

        Args:
            response (EquationExtractionResponse): The extracted equations and their metadata
            path (str): Directory path where the report should be saved
        """
        create_report(response, path)

    def analyze_equations(
        self,
        file_path: Optional[str] = None,
        url_path: Optional[str] = None,
        parsed_paper: Optional[ParseResponse] = None,
    ) -> Optional[EquationProcessingResponse]:
        """Analyze and extract equations from a scientific document.

        This method supports three input methods:
        1. Local PDF file path
        2. URL to a PDF (with special handling for arXiv URLs)
        3. Pre-parsed paper data

        Args:
            file_path (Optional[str]): Path to a local PDF file
            url_path (Optional[str]): URL to a PDF file (supports arXiv links)
            parsed_paper (Optional[ParseResponse]): Pre-parsed paper data

        Returns:
            Optional[EquationExtractionResponse]: Extracted equations and their metadata.
            Returns None if no valid input is provided.

        Examples:
            # From local file
            client.analyze_equations(file_path="path/to/paper.pdf")

            # From URL
            client.analyze_equations(url_path="https://arxiv.org/pdf/2203.00001.pdf")

            # From parsed paper
            client.analyze_equations(parsed_paper=parsed_data)
        """
        if file_path:
            with open(file_path, "rb") as pdf_file:
                parsed_document = self._ax_client.document.parse(file=pdf_file)
                print("We are almost there")
                response = self._ax_client.document.equation.process(
                    markdown=parsed_document.markdown,
                    interline_equations=parsed_document.interline_equations,
                    inline_equations=parsed_document.inline_equations,
                )

        elif url_path:
            if "arxiv" in url_path and "abs" in url_path:
                url_path = url_path.replace("abs", "pdf")
            url_file = requests.get(url_path)
            from io import BytesIO

            pdf_stream = BytesIO(url_file.content)
            parsed_document = self._ax_client.document.parse(file=url_file.content)
            print("We are almost there")
            response = self._ax_client.document.equation.process(
                markdown=parsed_document.markdown,
                interline_equations=parsed_document.interline_equations,
                inline_equations=parsed_document.inline_equations,
            )

        elif parsed_paper:
            response = EquationProcessingResponse.model_validate(
                self._ax_client.document.equation.process(**parsed_paper.model_dump()).model_dump()
            )

        else:
            print("Please provide either a file path or a URL to analyze.")
            return None

        return response

    def validate_equations(
        self,
        requirements: Sequence[VariableRequirement],
        loaded_equations: EquationProcessingResponse,
        include_internal_model: bool = False,
    ) -> EquationValidationResult:
        """Validate equations against a set of variable requirements.

        Args:
            requirements: List of variable requirements to validate
            loaded_equations: Previously processed equations to validate
            show_hypergraph: Whether to display the validation results graph (default: True)
            include_internal_model: Whether to include internal model equations in validation (default: False)

        Returns:
            EquationValidationResult containing the validation results
        """
        # equations_dict = loaded_equations.model_dump() if hasattr(loaded_equations, 'model_dump') else loaded_equations.dict()

        api_response = self._ax_client.document.equation.validate(
            variables=requirements, paper_equations=loaded_equations, include_internal_model=include_internal_model
        )

        return api_response

    def display_full_results(self, api_response: EquationValidationResult, user_choice):
        display_full_results(api_response, user_choice)

    def set_numerical_requirements(self, extracted_equations: EquationProcessingResponse):
        """Launch an interactive interface for setting numerical requirements for equations.

        This method opens an interactive table interface where users can specify
        requirements for variables found in the extracted equations.

        Args:
            extracted_equations: The equations to set requirements for

        Returns:
            The requirements set through the interactive interface
        """
        from .axtract.interactive_table import interactive_table

        result = interactive_table(extracted_equations)
        return result


class ToolsHelper:
    _ax_client: Axiomatic

    def __init__(self, ax_client: Axiomatic):
        self._ax_client = ax_client

    def tool_exec(self, tool: str, code: str, poll_interval: int = 3, debug: bool = True):
        """
        Helper function to schedule code execution for a specific tool and wait
        the execution to finish and return the output or error trace
        """
        if not tool.strip():
            return "Please specify a tool"
        else:
            tool_name = tool.strip()
            code_string = code

            tool_result = self._ax_client.tools.schedule(
                tool_name=tool_name,
                code=code_string,
            )
            if tool_result.is_success is True:
                job_id = str(tool_result.job_id)
                result = self._ax_client.tools.status(job_id=job_id)
                if debug:
                    print(f"job_id: {job_id}")
                while True:
                    result = self._ax_client.tools.status(job_id=job_id)
                    if result.status == "PENDING" or result.status == "RUNNING":
                        if debug:
                            print(f"status: {result.status}")
                        time.sleep(poll_interval)
                    else:
                        if debug:
                            print(f"status: {result.status}")
                        if result.status == "SUCCEEDED":
                            output = json.loads(result.output or "{}")
                            if not output["objects"]:
                                return result.output
                            else:
                                return {
                                    "job_id": job_id,
                                    "messages": output["messages"],
                                    "objects": self._load_objects_from_base64(output["objects"]),
                                }
                        else:
                            return result.error_trace
            else:
                return tool_result.error_trace

    def load(self, job_id: str, obj_key: str):
        result = self._ax_client.tools.status(job_id=job_id)
        if result.status == "SUCCEEDED":
            output = json.loads(result.output or "{}")
            if not output["objects"]:
                return result.output
            else:
                return self._load_objects_from_base64(output["objects"])[obj_key]
        else:
            return result.error_trace

    def _load_objects_from_base64(self, encoded_dict):
        loaded_objects = {}
        for key, encoded_str in encoded_dict.items():
            try:
                decoded_bytes = base64.b64decode(encoded_str)
                loaded_obj = dill.loads(decoded_bytes)
                loaded_objects[key] = loaded_obj
            except Exception as e:
                print(f"Error loading object for key '{key}': {e}")
                loaded_objects[key] = None
        return loaded_objects


class AsyncAxiomatic(AsyncBaseClient): ...



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .contour_cv2 import ContourCv2
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class AxesInfo(UniversalBaseModel):
    """
    Pydantic class for Axes Info
    """

    origin: typing.Optional[typing.List[typing.Optional[typing.Any]]] = None
    x_axis_len: typing.Optional[float] = None
    y_axis_len: typing.Optional[float] = None
    x_contour: typing.Optional[typing.List[typing.List[float]]] = None
    y_contour: typing.Optional[typing.List[typing.List[float]]] = None
    x_tick_marks: typing.Optional[typing.List[ContourCv2]] = None
    y_tick_marks: typing.Optional[typing.List[ContourCv2]] = None
    x_tick_vals: typing.Optional[typing.List[typing.List[typing.Optional[typing.Any]]]] = None
    y_tick_vals: typing.Optional[typing.List[typing.List[typing.Optional[typing.Any]]]] = None
    x_tick_mapping: typing.Optional[typing.Dict[str, typing.Optional[ContourCv2]]] = None
    y_tick_mapping: typing.Optional[typing.Dict[str, typing.Optional[ContourCv2]]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from . import circuit, component, document
from .circuit import Settings, Statement

__all__ = ["Settings", "Statement", "circuit", "component", "document"]



from typing import List
from ..types.response_equation import ResponseEquation
from pyvis.network import Network  # type: ignore


def normalize_latex_symbol(symbol: str) -> str:
    """
    Normalizes LaTeX symbols by removing unnecessary outer braces.
    For example: '{\vec{r}}' and '\vec{r}' will both become '\vec{r}'

    Args:
        symbol: LaTeX symbol string
    Returns:
        Normalized symbol string
    """
    # Remove outer braces if they exist
    symbol = symbol.strip()
    while symbol.startswith("{") and symbol.endswith("}"):
        symbol = symbol[1:-1].strip()
    return symbol


def generate_relation_graph(equations: List[ResponseEquation]) -> str:
    """
    Generates HTML code for a bipartite graph visualization.
    Green nodes represent equations, red nodes represent variables.

    Args:
        equations: List of EquationExtraction objects

    Returns:
        HTML string containing the graph visualization code
    """

    # Create a new network
    net = Network(height="900px", width="100%", bgcolor="#ffffff", font_color="#000000", notebook=False)

    # Track all variables to avoid duplicates
    all_variables = set()

    # Add equation nodes (green) and variable nodes (red)
    for eq in equations:
        # Add equation node with unique identifier
        eq_name = f"Eq: {eq.name}"  # Add ID to make each node unique
        net.add_node(
            eq_name,
            label=eq.name,
            color="#90EE90",  # Light green for equations
            shape="box",
            size=30,
            font=dict(
                size=14,
                color="#000000",  # Black text for better contrast on light green
            ),
            title=f"{eq.name}\n{eq.original_format}\n{eq.description}",
        )

        # Add variable nodes and edges
        for symbol in eq.latex_symbols:
            # Normalize the variable name
            var_name = normalize_latex_symbol(symbol.key)
            if var_name not in all_variables:
                net.add_node(
                    var_name,
                    label=var_name,
                    color="#E74C3C",  # Red for variables
                    shape="dot",
                    size=20,
                    font=dict(size=16, color="#000000"),
                    title=f"{var_name}: {symbol.value}",
                )
                all_variables.add(var_name)

            # Add edge between equation and variable
            net.add_edge(eq_name, var_name, color="#7F8C8D", width=1.5)

    # Configure physics for better layout
    net.set_options("""
    {
      "physics": {
        "barnesHut": {
          "gravitationalConstant": -2000,
          "centralGravity": 0.3,
          "springLength": 150,
          "springConstant": 0.04,
          "damping": 0.09,
          "avoidOverlap": 0.1
        },
        "minVelocity": 0.75
      }
    }
    """)

    # Generate a complete standalone HTML page
    html = net.generate_html()

    # Add a title and back button to the HTML
    html = html.replace(
        "<head>",
        """<head>
    <title>Equation Relation Graph</title>
    <style>
        .back-button {
            position: fixed;
            top: 20px;
            left: 20px;
            padding: 10px 15px;
            background-color: #4CAF50;
            color: white;
            border: none;
            border-radius: 5px;
            cursor: pointer;
            font-size: 16px;
            z-index: 1000;
        }
        .back-button:hover {
            background-color: #45a049;
        }
    </style>""",
    )

    html = html.replace(
        "<body>",
        """<body>
    <button class="back-button" onclick="window.close()">Close Graph</button>""",
    )

    return html



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class PlotInfo(UniversalBaseModel):
    """
    Pydantic class for Plot information to extract
    """

    x_axis_name: typing.Optional[str] = None
    x_axis_min: typing.Optional[float] = None
    x_axis_max: typing.Optional[float] = None
    x_axis_tick_values: typing.Optional[typing.List[float]] = None
    x_axis_unit: typing.Optional[str] = None
    x_scale: typing.Optional[str] = None
    y_axis_name: typing.Optional[str] = None
    y_axis_min: typing.Optional[float] = None
    y_axis_max: typing.Optional[float] = None
    y_axis_tick_values: typing.Optional[typing.List[float]] = None
    y_axis_unit: typing.Optional[str] = None
    y_scale: typing.Optional[str] = None
    num_of_lines: typing.Optional[int] = None
    black_gray_line: typing.Optional[bool] = None
    grid_lines: typing.Optional[bool] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from ...types.formalize_response import FormalizeResponse
from .types.solution_body_values_value import SolutionBodyValuesValue
from ...core.request_options import RequestOptions
from ...types.solution_response import SolutionResponse
from ...core.serialization import convert_and_respect_annotation_metadata
from ...core.pydantic_utilities import parse_obj_as
from ...errors.unprocessable_entity_error import UnprocessableEntityError
from ...types.http_validation_error import HttpValidationError
from json.decoder import JSONDecodeError
from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class SolutionClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def find(
        self,
        *,
        constraints: FormalizeResponse,
        values: typing.Dict[str, SolutionBodyValuesValue],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> SolutionResponse:
        """
        Finds a solution to a set of constraints provided partial values

        Parameters
        ----------
        constraints : FormalizeResponse

        values : typing.Dict[str, SolutionBodyValuesValue]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        SolutionResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic, FormalizeResponse

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.formalization.solution.find(
            constraints=FormalizeResponse(
                variables={"key": "value"},
                expressions=[],
            ),
            values={"key": 1},
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "formalization/solution/find",
            method="POST",
            json={
                "constraints": convert_and_respect_annotation_metadata(
                    object_=constraints, annotation=FormalizeResponse, direction="write"
                ),
                "values": convert_and_respect_annotation_metadata(
                    object_=values, annotation=typing.Dict[str, SolutionBodyValuesValue], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    SolutionResponse,
                    parse_obj_as(
                        type_=SolutionResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncSolutionClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def find(
        self,
        *,
        constraints: FormalizeResponse,
        values: typing.Dict[str, SolutionBodyValuesValue],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> SolutionResponse:
        """
        Finds a solution to a set of constraints provided partial values

        Parameters
        ----------
        constraints : FormalizeResponse

        values : typing.Dict[str, SolutionBodyValuesValue]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        SolutionResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic, FormalizeResponse

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.formalization.solution.find(
                constraints=FormalizeResponse(
                    variables={"key": "value"},
                    expressions=[],
                ),
                values={"key": 1},
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "formalization/solution/find",
            method="POST",
            json={
                "constraints": convert_and_respect_annotation_metadata(
                    object_=constraints, annotation=FormalizeResponse, direction="write"
                ),
                "values": convert_and_respect_annotation_metadata(
                    object_=values, annotation=typing.Dict[str, SolutionBodyValuesValue], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    SolutionResponse,
                    parse_obj_as(
                        type_=SolutionResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import pydantic
import typing
from .dict_item import DictItem
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class ResponseEquation(UniversalBaseModel):
    """
    Pydantic model representing a scientific or mathematical equation.
    """

    name: str = pydantic.Field()
    """
    Human-readable name of the equation (e.g., 'Newton's Law')
    """

    description: str = pydantic.Field()
    """
    Text describing the equation
    """

    original_format: str = pydantic.Field()
    """
    Original form of the equation (e.g., LaTeX)
    """

    wolfram_expressions: str = pydantic.Field()
    """
    Equation in Wolfram Language (or other symbolic form)
    """

    latex_symbols: typing.List[DictItem] = pydantic.Field()
    """
    Detailed metadata for each latex variable in original format
    """

    wolfram_symbols: typing.List[str] = pydantic.Field()
    """
    List of symbols used in the wolfram expression
    """

    narrative_assumptions: typing.List[str] = pydantic.Field()
    """
    Narrative text describing assumptions/approximations
    """

    type: typing.List[str] = pydantic.Field()
    """
    List of equation classifications (e.g., ['scalar','tensor'])
    """

    field_tags: typing.List[str] = pydantic.Field()
    """
    Classification tags (e.g., ['quantum mechanics','gravitation'])
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.client_wrapper import SyncClientWrapper
from .document.client import DocumentClient
from .circuit.client import CircuitClient
from .component.client import ComponentClient
from ..core.client_wrapper import AsyncClientWrapper
from .document.client import AsyncDocumentClient
from .circuit.client import AsyncCircuitClient
from .component.client import AsyncComponentClient


class PicClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper
        self.document = DocumentClient(client_wrapper=self._client_wrapper)
        self.circuit = CircuitClient(client_wrapper=self._client_wrapper)
        self.component = ComponentClient(client_wrapper=self._client_wrapper)


class AsyncPicClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper
        self.document = AsyncDocumentClient(client_wrapper=self._client_wrapper)
        self.circuit = AsyncCircuitClient(client_wrapper=self._client_wrapper)
        self.component = AsyncComponentClient(client_wrapper=self._client_wrapper)
